"""
DeepLabCut2.0 Toolbox (deeplabcut.org)
© A. & M. Mathis Labs
https://github.com/AlexEMG/DeepLabCut
Please see AUTHORS for contributors.

https://github.com/AlexEMG/DeepLabCut/blob/master/AUTHORS
Licensed under GNU Lesser General Public License v3.0

"""

from deeplabcut.gui.analyze_videos import *
from deeplabcut.gui.create_new_project import *
from deeplabcut.gui.create_training_dataset import *
from deeplabcut.gui.evaluate_network import *
from deeplabcut.gui.extract_frames import *
from deeplabcut.gui.extract_outlier_frames import *
from deeplabcut.gui.label_frames import *
from deeplabcut.gui.load_project import *
from deeplabcut.gui.refine_labels import *
from deeplabcut.gui.train_network import *
from deeplabcut.gui.welcome import *
from deeplabcut.gui.media import *
from deeplabcut.gui.create_videos import *